@@ -8,15 +8,19 @@
 #
 
 import ast
+import asyncio
 import operator
 import os
-import re
 import tarfile
+from traceback import format_exc
 from typing import Dict
 from typing import Generator
+from typing import Iterable
 from typing import List
+from typing import Mapping
 from typing import NamedTuple
 from typing import Tuple
+from typing import Type
 from typing import Union
 from zipfile import ZipFile
 
@@ -36,12 +40,13 @@ from _packagedcode.pypi import PypiWheelHandler
 from _packagedcode.pypi import PythonSetupPyHandler
 from _packagedcode.pypi import SetupCfgHandler
 from _packagedcode.pypi import can_process_dependent_package
+from python_inspector import pyinspector_settings as settings
 from python_inspector import utils_pypi
 from python_inspector.error import NoVersionsFound
 from python_inspector.setup_py_live_eval import iter_requirements
 from python_inspector.utils import Candidate
 from python_inspector.utils import contain_string
-from python_inspector.utils import get_response
+from python_inspector.utils import get_response_async
 from python_inspector.utils_pypi import PypiSimpleRepository
 
 
@@ -71,7 +76,7 @@ class Result(NamedTuple):
 
 
 def get_requirements_from_distribution(
-    handler: BasePypiHandler,
+    handler: Type[BasePypiHandler],
     location: str,
 ) -> List[Requirement]:
     """
@@ -83,9 +88,15 @@ def get_requirements_from_distribution(
     if not os.path.exists(location):
         return []
     reqs = []
-    for package_data in handler.parse(location):
-        dependencies = package_data.dependencies
-    reqs.extend(get_requirements_from_dependencies(dependencies=dependencies))
+    try:
+        for package_data in handler.parse(location):
+            dependencies = package_data.dependencies
+            reqs.extend(get_requirements_from_dependencies(dependencies=dependencies))
+    except Exception as e:
+        trace = format_exc()
+        raise Exception(
+            f"Failed to get_requirements_from_distribution for: {location!r}\n{trace}"
+        ) from e
     return reqs
 
 
@@ -161,7 +172,9 @@ def is_valid_version(
     """
     if parsed_version in bad_versions:
         return False
-    if any(parsed_version not in r.specifier for r in requirements[identifier]):
+    if any(
+        not r.specifier.contains(parsed_version, prereleases=True) for r in requirements[identifier]
+    ):
         if all(not r.specifier for r in requirements[identifier]):
             return True
         return False
@@ -170,6 +183,8 @@ def is_valid_version(
 
 def get_python_version_from_env_tag(python_version: str) -> str:
     """
+    Return the python version extracted from an environment tag.
+
     >>> assert get_python_version_from_env_tag("310") == "3.10"
     >>> assert get_python_version_from_env_tag("39") == "3.9"
     """
@@ -179,7 +194,7 @@ def get_python_version_from_env_tag(python_version: str) -> str:
     return python_version
 
 
-def fetch_and_extract_sdist(
+async def fetch_and_extract_sdist(
     repos: List[PypiSimpleRepository], candidate: Candidate, python_version: str
 ) -> Union[str, None]:
     """
@@ -192,7 +207,7 @@ def fetch_and_extract_sdist(
     the required ``python_version``.
     Raise an Exception if extraction fails.
     """
-    sdist = utils_pypi.download_sdist(
+    sdist = await utils_pypi.download_sdist(
         name=candidate.name,
         version=str(candidate.version),
         repos=repos,
@@ -208,21 +223,21 @@ def fetch_and_extract_sdist(
 def get_sdist_file_path_from_filename(sdist):
     if sdist.endswith(".tar.gz"):
         sdist_file = sdist.rstrip(".tar.gz")
-        with tarfile.open(os.path.join(utils_pypi.CACHE_THIRDPARTY_DIR, sdist)) as file:
+        with tarfile.open(os.path.join(settings.CACHE_THIRDPARTY_DIR, sdist)) as file:
             file.extractall(
-                os.path.join(utils_pypi.CACHE_THIRDPARTY_DIR, "extracted_sdists", sdist_file)
+                os.path.join(settings.CACHE_THIRDPARTY_DIR, "extracted_sdists", sdist_file)
             )
     elif sdist.endswith(".zip"):
         sdist_file = sdist.rstrip(".zip")
-        with ZipFile(os.path.join(utils_pypi.CACHE_THIRDPARTY_DIR, sdist)) as zip:
+        with ZipFile(os.path.join(settings.CACHE_THIRDPARTY_DIR, sdist)) as zip:
             zip.extractall(
-                os.path.join(utils_pypi.CACHE_THIRDPARTY_DIR, "extracted_sdists", sdist_file)
+                os.path.join(settings.CACHE_THIRDPARTY_DIR, "extracted_sdists", sdist_file)
             )
 
     else:
         raise Exception(f"Unable to extract sdist {sdist}")
 
-    return os.path.join(utils_pypi.CACHE_THIRDPARTY_DIR, "extracted_sdists", sdist_file, sdist_file)
+    return os.path.join(settings.CACHE_THIRDPARTY_DIR, "extracted_sdists", sdist_file, sdist_file)
 
 
 def get_requirements_from_dependencies(
@@ -255,7 +270,7 @@ def remove_extras(identifier: str) -> str:
     return name
 
 
-def get_reqs_from_requirements_file_in_sdist(sdist_location: str, files: str) -> List[Requirement]:
+def get_reqs_from_requirements_file_in_sdist(sdist_location: str, files: List) -> List[Requirement]:
     """
     Return a list of parsed requirements from the ``sdist_location`` sdist location
     """
@@ -283,7 +298,7 @@ def get_requirements_from_python_manifest(
     """
     Return a list of parsed requirements from the ``sdist_location`` sdist location
     """
-    # Look in requirements file if and only if they are refered in setup.py or setup.cfg
+    # Look in requirements file if and only if they are referred in setup.py or setup.cfg
     # And no deps have been yielded by requirements file.
     requirements = list(
         get_reqs_from_requirements_file_in_sdist(
@@ -316,10 +331,6 @@ def get_requirements_from_python_manifest(
                         and elem.value.func.id == "setup"
                     )
                 ]
-                if len(setup_fct) == 0:
-                    raise Exception(
-                        f"Unable to collect setup.py dependencies securely: {setup_py_location}"
-                    )
                 if len(setup_fct) > 1:
                     print(
                         f"Warning: identified multiple definitions of 'setup()' in {setup_py_location}, "
@@ -329,20 +340,17 @@ def get_requirements_from_python_manifest(
                 install_requires = [
                     k.value for k in setup_fct.value.keywords if k.arg == "install_requires"
                 ]
-                if len(install_requires) == 0:
-                    raise Exception(
-                        f"Unable to collect setup.py dependencies securely: {setup_py_location}"
-                    )
-                if len(install_requires) > 1:
-                    print(
-                        f"Warning: identified multiple definitions of 'install_requires' in "
-                        "{setup_py_location}, defaulting to the first occurrence"
-                    )
-                install_requires = install_requires[0].elts
-                if len(install_requires) != 0:
-                    raise Exception(
-                        f"Unable to collect setup.py dependencies securely: {setup_py_location}"
-                    )
+                if install_requires:
+                    if len(install_requires) > 1:
+                        print(
+                            f"Warning: identified multiple definitions of 'install_requires' in "
+                            "{setup_py_location}, defaulting to the first occurrence"
+                        )
+                    install_requires = install_requires[0].elts
+                    if len(install_requires) != 0:
+                        raise Exception(
+                            f"Unable to collect setup.py dependencies securely: {setup_py_location}"
+                        )
 
 
 DEFAULT_ENVIRONMENT = utils_pypi.Environment.from_pyver_and_os(
@@ -361,7 +369,7 @@ class PythonInputProvider(AbstractProvider):
         self.environment = environment
         self.environment_marker = get_environment_marker_from_environment(self.environment)
         self.repos = repos or []
-        self.versions_by_package = {}
+        self.versions_by_package: Dict[str, List[Version]] = {}
         self.dependencies_by_purl = {}
         self.wheel_or_sdist_by_package = {}
         self.analyze_setup_py_insecurely = analyze_setup_py_insecurely
@@ -388,25 +396,38 @@ class PythonInputProvider(AbstractProvider):
         transitive = all(p is not None for _, p in information[identifier])
         return transitive, identifier
 
-    def get_versions_for_package(
-        self, name: str, repo: Union[List[PypiSimpleRepository], None] = None
-    ) -> List[Version]:
+    def get_versions_for_package(self, name: str) -> List[Version]:
         """
         Return a list of versions for a package.
         """
-        if repo and self.environment:
-            return self.get_versions_for_package_from_repo(name, repo)
+        versions = self.versions_by_package.get(name)
+        if not versions:
+            return asyncio.run(self.fill_versions_for_package(name))
         else:
-            return self.get_versions_for_package_from_pypi_json_api(name)
+            return versions
 
-    def get_versions_for_package_from_repo(
+    async def fill_versions_for_package(self, name: str) -> List[Version]:
+        versions = self.versions_by_package.get(name) or []
+        if versions:
+            return versions
+
+        if self.repos and self.environment:
+            for repo in self.repos:
+                versions.extend(await self._get_versions_for_package_from_repo(name, repo))
+        else:
+            versions.extend(await self._get_versions_for_package_from_pypi_json_api(name))
+
+        self.versions_by_package[name] = versions
+        return versions
+
+    async def _get_versions_for_package_from_repo(
         self, name: str, repo: PypiSimpleRepository
     ) -> List[Version]:
         """
         Return a list of versions for a package name from a repo
         """
         versions = []
-        for version, package in repo.get_package_versions(name).items():
+        for version, package in (await repo.get_package_versions(name)).items():
             python_version = parse_version(
                 get_python_version_from_env_tag(python_version=self.environment.python_version)
             )
@@ -425,34 +446,50 @@ class PythonInputProvider(AbstractProvider):
                 )
             if valid_wheel_present or pypi_valid_python_version:
                 versions.append(version)
+
         return versions
 
-    def get_versions_for_package_from_pypi_json_api(self, name: str) -> List[Version]:
+    async def _get_versions_for_package_from_pypi_json_api(self, name: str) -> List[Version]:
         """
         Return a list of versions for a package name from the PyPI.org JSON API
         """
-        if name not in self.versions_by_package:
-            api_url = f"https://pypi.org/pypi/{name}/json"
-            resp = get_response(api_url)
-            if not resp:
-                self.versions_by_package[name] = []
-            releases = resp.get("releases") or {}
-            self.versions_by_package[name] = releases.keys() or []
-        versions = self.versions_by_package[name]
-        return versions
+        api_url = f"https://pypi.org/pypi/{name}/json"
+        resp = await get_response_async(api_url)
+        if not resp:
+            return []
+        releases = resp.get("releases") or {}
+        return releases.keys() or []
 
     def get_requirements_for_package(
         self, purl: PackageURL, candidate: Candidate
-    ) -> Generator[Requirement, None, None]:
+    ) -> List[Requirement]:
+        dependencies = self.dependencies_by_purl.get(str(purl))
+        if not dependencies:
+            return asyncio.run(self.fill_requirements_for_package(purl, candidate))
+        else:
+            return dependencies
+
+    async def fill_requirements_for_package(
+        self, purl: PackageURL, candidate: Candidate
+    ) -> List[Requirement]:
         """
         Yield requirements for a package.
         """
-        if self.repos and self.environment:
-            return self.get_requirements_for_package_from_pypi_simple(candidate)
-        else:
-            return self.get_requirements_for_package_from_pypi_json_api(purl)
+        dependencies = self.dependencies_by_purl.get(str(purl)) or []
+        if dependencies:
+            return dependencies
 
-    def get_requirements_for_package_from_pypi_simple(
+        if self.repos and self.environment:
+            dependencies.extend(
+                await self._get_requirements_for_package_from_pypi_simple(candidate)
+            )
+        else:
+            dependencies.extend(await self._get_requirements_for_package_from_pypi_json_api(purl))
+
+        self.dependencies_by_purl[str(purl)] = dependencies
+        return dependencies
+
+    async def _get_requirements_for_package_from_pypi_simple(
         self, candidate: Candidate
     ) -> List[Requirement]:
         """
@@ -462,7 +499,7 @@ class PythonInputProvider(AbstractProvider):
             get_python_version_from_env_tag(python_version=self.environment.python_version)
         )
 
-        wheels = utils_pypi.download_wheel(
+        wheels = await utils_pypi.download_wheel(
             name=candidate.name,
             version=str(candidate.version),
             environment=self.environment,
@@ -472,22 +509,19 @@ class PythonInputProvider(AbstractProvider):
 
         if wheels:
             for wheel in wheels:
-                wheel_location = os.path.join(utils_pypi.CACHE_THIRDPARTY_DIR, wheel)
+                wheel_location = os.path.join(settings.CACHE_THIRDPARTY_DIR, wheel)
                 requirements = get_requirements_from_distribution(
                     handler=PypiWheelHandler,
                     location=wheel_location,
                 )
-                if requirements:
-                    yield from requirements
                 # We are only looking at the first wheel and not other wheels
-                break
-
+                return requirements
         else:
-            sdist_location = fetch_and_extract_sdist(
+            sdist_location = await fetch_and_extract_sdist(
                 repos=self.repos, candidate=candidate, python_version=python_version
             )
             if not sdist_location:
-                return
+                return []
 
             setup_py_location = os.path.join(
                 sdist_location,
@@ -499,9 +533,7 @@ class PythonInputProvider(AbstractProvider):
             )
 
             if self.analyze_setup_py_insecurely:
-                yield from get_reqs_insecurely(
-                    setup_py_location=setup_py_location,
-                )
+                return get_reqs_insecurely(setup_py_location=setup_py_location)
             else:
                 requirements = list(
                     get_setup_requirements(
@@ -511,35 +543,31 @@ class PythonInputProvider(AbstractProvider):
                     )
                 )
                 if requirements:
-                    yield from requirements
+                    return requirements
                 else:
-                    # Look in requirements file if and only if thy are refered in setup.py or setup.cfg
+                    # Look in requirements file if and only if thy are referred in setup.py or setup.cfg
                     # And no deps have been yielded by requirements file
-
-                    yield from get_requirements_from_python_manifest(
+                    return get_requirements_from_python_manifest(
                         sdist_location=sdist_location,
                         setup_py_location=setup_py_location,
                         files=[setup_cfg_location, setup_py_location],
                         analyze_setup_py_insecurely=self.analyze_setup_py_insecurely,
                     )
 
-    def get_requirements_for_package_from_pypi_json_api(
+    async def _get_requirements_for_package_from_pypi_json_api(
         self, purl: PackageURL
     ) -> List[Requirement]:
         """
         Return requirements for a package from the PyPI.org JSON API
         """
         # if no repos are provided use the incorrect but fast JSON API
-        if str(purl) not in self.dependencies_by_purl:
-            api_url = f"https://pypi.org/pypi/{purl.name}/{purl.version}/json"
-            resp = get_response(api_url)
-            if not resp:
-                self.dependencies_by_purl[str(purl)] = []
-            info = resp.get("info") or {}
-            requires_dist = info.get("requires_dist") or []
-            self.dependencies_by_purl[str(purl)] = requires_dist
-        for dependency in self.dependencies_by_purl[str(purl)]:
-            yield Requirement(dependency)
+        api_url = f"https://pypi.org/pypi/{purl.name}/{purl.version}/json"
+        resp = await get_response_async(api_url)
+        if not resp:
+            return []
+        info = resp.get("info") or {}
+        requires_dist = info.get("requires_dist") or []
+        return list(map(lambda r: Requirement(r), requires_dist))
 
     def get_candidates(
         self,
@@ -549,7 +577,7 @@ class PythonInputProvider(AbstractProvider):
         bad_versions: List[str],
         name: str,
         extras: Dict,
-    ) -> Generator[Candidate, None, None]:
+    ) -> Iterable[Candidate]:
         """
         Generate candidates for the given identifier. Overridden.
         """
@@ -571,21 +599,17 @@ class PythonInputProvider(AbstractProvider):
     def _iter_matches(
         self,
         identifier: str,
-        requirements: Dict,
+        requirements: Mapping[str, List[Requirement]],
         incompatibilities: Dict,
     ) -> Generator[Candidate, None, None]:
         """
-        Yield candidates for the given identifier, requirements and incompatibilities
+        Yield candidates for the given identifier, requirements and incompatibilities.
         """
         name = remove_extras(identifier=identifier)
         bad_versions = {c.version for c in incompatibilities[identifier]}
         extras = {e for r in requirements[identifier] for e in r.extras}
         versions = []
-        if not self.repos:
-            versions.extend(self.get_versions_for_package(name=name))
-        else:
-            for repo in self.repos:
-                versions.extend(self.get_versions_for_package(name=name, repo=repo))
+        versions.extend(self.get_versions_for_package(name=name))
 
         if not versions:
             if self.ignore_errors:
@@ -605,7 +629,7 @@ class PythonInputProvider(AbstractProvider):
     def find_matches(
         self,
         identifier: str,
-        requirements: List[Requirement],
+        requirements: Mapping[str, List[Requirement]],
         incompatibilities: Dict,
     ) -> List[Candidate]:
         """Find all possible candidates that satisfy given constraints. Overridden."""
@@ -624,7 +648,7 @@ class PythonInputProvider(AbstractProvider):
             return True
         return False
 
-    def _iter_dependencies(self, candidate: Candidate) -> Generator[Requirement, None, None]:
+    def _iter_dependencies(self, candidate: Candidate) -> Iterable[Requirement]:
         """
         Yield dependencies for the given candidate.
         """
