@@ -155,6 +155,7 @@ PLATFORMS_BY_OS = {
         "manylinux2014_x86_64",
         "manylinux_2_27_x86_64",
         "manylinux_2_28_x86_64",
+        "manylinux_2_31_x86_64",
         "manylinux2014_aarch6",
         "musllinux_1_2_x86_64",
         "manylinux_2_33_aarch64",
@@ -1806,7 +1807,7 @@ async def get_remote_file_content(
         if login and password:
             auth = aiohttp.BasicAuth(login, password)
 
-    async with aiohttp.ClientSession() as session:
+    async with aiohttp.ClientSession(trust_env=True) as session:
         async with session.get(url, allow_redirects=True, headers=headers, auth=auth) as response:
             status = response.status
             if status != requests.codes.ok:  # NOQA
