@@ -20,6 +20,7 @@ from click.testing import CliRunner
 from commoncode.testcase import FileDrivenTesting
 
 from python_inspector.resolve_cli import resolve_dependencies
+from python_inspector.resolve_cli import __version__
 
 # Used for tests to regenerate fixtures with regen=True
 REGEN_TEST_FIXTURES = os.getenv("PYINSP_REGEN_TEST_FIXTURES", False)
@@ -414,6 +415,17 @@ def test_cli_with_pinned_requirements_file():
     )
 
 
+@pytest.mark.online
+def test_cli_with_hash_requirements():
+    requirements_file = test_env.get_test_loc("hash-requirements.txt")
+    expected_file = test_env.get_test_loc("hash-requirements.txt-expected.json", must_exist=False)
+    check_requirements_resolution(
+        requirements_file=requirements_file,
+        expected_file=expected_file,
+        regen=REGEN_TEST_FIXTURES,
+    )
+
+
 @pytest.mark.online
 def test_cli_with_setup_py_failure():
     setup_py_file = setup_test_env.get_test_loc("simple-setup.py")
@@ -533,7 +545,7 @@ def test_passing_of_json_pdt_and_json_flags():
 def test_version_option():
     options = ["--version"]
     rc, stdout, stderr = run_cli(options=options)
-    assert "0.14.3" in stdout
+    assert __version__ in stdout
 
 
 def test_passing_of_netrc_file_that_does_not_exist():
